"""
This module contains functionality for accessing the realtime parts of
the PacketLogic system, such as system diagnostics, channel statistics
and dynamic netobject items.

Do NOT import this module directly. The L{PLConnection
<packetlogic2.PLConnection>} object should be used, e.g:

    >>> import packetlogic2
    >>> pl = packetlogic2.connect("192.168.1.25", "admin", "pldemo00")
    >>> rt = pl.Realtime()

@var __revision__: Revision identifier.
"""

__revision__ = '%K%'

from thisversion import verify_ip, negotiate_flicka_with_timeout
from thisversion import PL
from thisversion import PLd

import socket
import select
import time
import sys

from packetlogic2.exceptions import PLDPermissionError, PLDNotAvailable, PLDUnsupportedFeatureError, PLDServerError, PLDUnsupportedInFirmware

class PLD(object):
	"""
	Class for connecting and communicating with packetlogicd,
	the realtime part of PacketLogic.

	When using the multiversion API you will access this class via the
	L{PLConnection.Realtime() <packetlogic2.PLConnection.Realtime>}
	method.

	All functions that communicate with the PacketLogic Realtime
	system might raise errors common to all or many of the
	functions.Here is a list of such errors, that might not be
	listed on each method:

	@raises ValueError: As the superclass of an internal BadMessageError,
	if a malformed message is recieved on the PacketLogic connection.
	That normally will never happen though.

	@raises Realtime.PLDPermissionError: If the user lacks permission
	        for the requested action.

	@raises IOError: If a generic I/O error occurs while reading or
	        writing on the PacketLogic Realtime connection.

	@group Dynamic NetObject items: dyn_add,dyn_list,dyn_list_full,dyn_list_mask,dyn_list_no,dyn_remove
	@group Firewall Log: fwlog_clear,fwlog_query_offsets,fwlog_query_rule,fwlog_query_time_range
	@group Volume Based Shaping: vbs_list,vbs_query,vbs_reset,vbs_set
	@group Internal: dump_bgptable,dump_drdlstats,inject_packet,reload_drdl
	@group Realtime Data: add_channels_callback,add_netobj_callback,add_service_callback,add_shapingcnt_callback,add_stats_callback,add_update_done_callback,get_channels_data,get_netobj_data,get_services_data,get_stats_data,shapingcnt_request_all,stop_updating,update_done_callbacks,update_forever
	
	"""
	
	SUPPORTS_TIMEOUT = True # Actually it is pldb.Ruleset that is checked for this flag, but it seems apropriate to set this too.
	
	def __init__(self, host, user, password, port=42001, timeout=60.0):
		"""
		Initialize and connect to packetlogicd.

		@type host: C{str}
		@param host: A hostname or IP address for the PacketLogic
		             system you want to connect to.

		@type user: C{str}
		@param user: The username you want to authenticate
		             with.  The system share users with the
		             client. You can create the user you want
		             to use with the client.

		@type password: C{str}
		@param password: The password to authenticate with.

		@type port: C{int}
		@param port: The default port is 42001 and normally
		             you don't have to change that.

		@type timeout: C{float} or C{None}
		@param timeout: The connection timeout in seconds. Default None.
			Set to None to use blocking mode during connect, with no specified timeout.
			(A system dependent timeout may or may not still occur after some time.)
			New in version v11.3beta3
		       
		@raises RuntimeError: If authorization fails
		@raises RuntimeError: If server and API version does not match
		@raises IOError: B{"Operation timed out"}: If operation times out without response.
		@raises IOError: If an error occures on the socket when attempting to connect.
		
		"""
		self.host = host
		self.port = port
		self.user = user
		self.timeout = timeout
		#self.timeout = timeout or 60.0 # FIXME Should we set this to the timeout argument too?
		self.password = password
		self.authed = False
		self.s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
		self.s.setblocking(False)
		self.s.settimeout(timeout)
		try:
			self.s.connect((host, port))
		except socket.error:
			typ, val, tb = sys.exc_info()
			raise IOError, val, tb
		self.io = PL.PLIO(self.s.fileno())
		self.handlers = {}
		self.stats_callbacks = []
		self.chans_callbacks = []
		self.no_callbacks = []
		self.update_done_callbacks = []
		self.service_callbacks = []
		self.interesting_netobjects = []
		self._update_once_data = None
		self._cache_stats = None
		self._cache_stats_ts = 0
		self._cache_channels = None
		self._cache_channels_ts = 0
		self._NO = RealtimeNetObjectTree()
		self._netobj_ts = 0
		self._features = None
		self._got_statsdata = False

		self._services = {}
		self._service_ts = 0

		self._strtbl = {}

		negotiate_flicka_with_timeout(self.s, self.io, timeout)

		self.handlers[PLd.PROTOCOL_SERVER_VERSION] = self._handle_vers
		self.handlers[PLd.PROTOCOL_SERVER_AUTHENTICATION_OK] = self._handle_auth
		self.handlers[PLd.PROTOCOL_SERVER_FEATURES] = self._handle_features
		self.handlers[PLd.PROTOCOL_SERVER_EPERM] = self._handle_eperm
		self.handlers[PLd.PROTOCOL_SERVER_NOT_AVAILABLE] = self._handle_eagain
		self.handlers[PLd.PROTOCOL_SERVER_STATS_ZONEINFO_BEGIN] = self._handle_statinfo_begin
		self.handlers[PLd.PROTOCOL_SERVER_STATS_ZONEINFO_END] = self._handle_statinfo_end
		self.handlers[PLd.PROTOCOL_SERVER_STATS_ZONEINFO] = self._handle_statinfo
		self.handlers[PLd.PROTOCOL_SERVER_CHANNEL_STATS_INFO] = self._handle_chaninfo
		self.handlers[PLd.PROTOCOL_SERVER_NO_RESET] = self._handle_netobj_reset
		self.handlers[PLd.PROTOCOL_SERVER_UPDATE_DONE] = self._handle_update_done
		self.handlers[PLd.PROTOCOL_SERVER_DYNAMICNETOBJECT_LIST] = self._handle_dynlist_err

		self._handler_add(PLd.PROTOCOL_SERVER_STRTABLE_BEGIN, self._handle_strtable_begin)
		self._handler_add(PLd.PROTOCOL_SERVER_STRTABLE_ENTRIES, self._handle_strtable_entries)
		self._handler_add(PLd.PROTOCOL_SERVER_STRTABLE_END, self._handle_strtable_end)

		self._handler_add(PLd.PROTOCOL_SERVER_SERVICE_ADD, self._handle_service_add)
		self._handler_add(PLd.PROTOCOL_SERVER_SERVICE_REMOVE, self._handle_service_remove)
		self._handler_add(PLd.PROTOCOL_SERVER_SERVICE_UPDATE, self._handle_service_update)


		PLd.doAuth(self.io, user, password)

		while not self.authed:
			if not self._process(self.timeout):
				raise IOError("Operation timed out")

		# In V11 the server will have NetObjects (realtime stats) visible at start, so send a Hide.
		PLd.netobjectsHide(self.io)

	def _handler_add(self, cmd, handler):
		self.handlers[cmd] = handler

	def _handler_remove(self, cmd):
		if self.handlers.has_key(cmd):
			del self.handlers[cmd]

	def _handle_auth(self, m):
		if PLd.handleAuth(m) == 0:
			raise RuntimeError("Authorization failed")
		self.authed = True

	def _handle_vers(self, m):
		ver = PLd.handleVersion(m)
		if ver != PLd.PROTOCOL_VERSION:
			raise RuntimeError("Bad server version (got %d expected %d)" % (ver, PLd.PROTOCOL_VERSION))
		self.protocolVersion = ver

	def _handle_features(self, m):
		class Features:
			def __repr__(self):
				return "<Features [%s]>" % (", ".join([n for n in dir(self) if not n.startswith('_') and getattr(self, n)]))
			
		bits = PLd.handleFeatures(m)
		features = Features()
		features.bgp = bool(bits & PLd.FEATURES_BGP)
		features.fw = bool(bits & PLd.FEATURES_FW)
		features.shaping = bool(bits & PLd.FEATURES_SHAPING)
		features.statistics = bool(bits & PLd.FEATURES_STATISTICS)
		features.connsearch = bool(bits & PLd.FEATURES_CONNSEARCH)
		features.enterprise = bool(bits & PLd.FEATURES_ENTERPRISE)
		features.monitor = bool(bits & PLd.FEATURES_MONITOR)
		features.divert = bool(bits & PLd.FEATURES_DIVERT)
		features.vbs = bool(bits & PLd.FEATURES_VBS)
		self._features = features

	def _handle_eperm(self, m):
		cmd = PLd.handleEperm(m)
		raise PLDPermissionError("Permission denied executing command (%x)" % cmd)

	def _handle_eagain(self, m):
		raise PLDNotAvailable("Data not available: '%s'" % PLd.handleNotAvail(m))

	def _handle_statinfo_end(self, m):
		def datahandler(m):
			PLd.handleStatsZoneData(m, self.sysdiag_zones)
			self._got_statsdata = True

		self._handler_add(PLd.PROTOCOL_SERVER_STATS_ZONEDATA, datahandler)

	def _handle_statinfo_begin(self, m):
		self._handler_remove(PLd.PROTOCOL_SERVER_STATS_ZONEDATA);
		self.sysdiag_zones = {}

	def _handle_statinfo(self, m):
		(zid, zone) = PLd.handleStatsZoneInfo(m)
		self.sysdiag_zones[zid] = zone;
		
	def _handle_chaninfo(self, m):
		chans = PLd.handleChannelsInfo(m)
		def datahandler(m):
			r = PLd.handleChannelsData(m, chans)
			for f in self.chans_callbacks:
				f(r)

		self._handler_add(PLd.PROTOCOL_SERVER_CHANNEL_STATS_DATA, datahandler)

	def _handle_dynlist_err(self, m):
		msg = PLd.handleDynList(m)
		raise PLDServerError("Error from PLD server on current command: '%s'" % msg)

	def _has_feature(self, feature):
		""" Return True or False depending on if the feature is enabled."""
		if self._features is None:
			raise RuntimeError, "Never recieved features message at connect."

		return getattr(self._features, feature)
		
	def _require_features(self, *features):
		""" If any of the listed features aren't enabled, a PLDUnsupportedFeatureError is raised."""
		for feature in features:
			if not self._has_feature(feature):
				raise PLDUnsupportedFeatureError
				
	def _flush(self):
		while self.io.wantWrite():
			outp = [self.s]

			(i, o, e) = select.select([], outp, outp, self.timeout)

			if len(i) + len(o) + len(e) == 0:
				raise IOError("Operation timed out")

			if len(e) > 0:
				raise IOError("Select returned error")
			
			if len(o) > 0:
				if not self.io.doWrite():
					raise IOError("Error writing to socket")

	def _handle_update_done(self, m):
		for f,under in self.no_callbacks:
			if not self.handlers.has_key(PLd.PROTOCOL_SERVER_NO_UPDATE):
				self._NO._link_parents()
				self._handler_add(PLd.PROTOCOL_SERVER_NO_UPDATE, self._handle_netobj_updateinfo)

			if under is not None:
				f(self._NO.list(under[:-1], recursive=False))
			else:
				f(self._NO)
		if self._got_statsdata:
			self._got_statsdata = False
			czones = dict(map(lambda q: (q[0], dict(map(lambda a: (a['name'], a), q[1]))), self.sysdiag_zones.values()))
			for f in self.stats_callbacks:
				f(czones)
		for f in self.service_callbacks:
			f(self._services)
			
		for c in self.update_done_callbacks:
			c()

	def _handle_strtable_begin(self, m):
		s, l = PLd.handleStrtblBegin(m)
		self._loading_strtbl_siz = s
		self._loading_strtbl = l
	def _handle_strtable_entries(self, m):
		self._loading_strtbl += PLd.handleStrtblEntries(m)
	def _handle_strtable_end(self, m):
		assert len(self._loading_strtbl) == self._loading_strtbl_siz
		self._strtbl = dict(self._loading_strtbl)
		self._loading_strtbl_siz = None
		self._loading_strtbl = None

	def _handle_netobj_reset(self, m):
		"""
		This is an internal function which converts NetObject Reset messages
		into Python data to be used by the NetObject callback function
		"""
		def datahandler(m):
			r = PLd.handleNOReset(m)
			self._NO._clear()

		self._handler_add(PLd.PROTOCOL_SERVER_NO_RESET, datahandler)
		self._handler_remove(PLd.PROTOCOL_SERVER_NO_UPDATE)
		self.handlers[PLd.PROTOCOL_SERVER_NO_RESET](m)

	def _handle_netobj_add(self, m):
		"""
		This is an internal function which converts NetObject Add messages
		into Python data to be used by the NetObject callback function
		"""
		def datahandler(m):
			nolist = PLd.handleNOAdd(m, RealtimeNetObject)
			self._NO._extend(nolist)
			for no in nolist:
				self._NO._link_parent_one(no)
				for noname in self.interesting_netobjects:
					fp = no.fullpath + "/"
					if noname.startswith(fp):
						PLd.netobjectShow(self.io, no.id)
						break

		self._handler_add(PLd.PROTOCOL_SERVER_NO_ADD, datahandler)
		self.handlers[PLd.PROTOCOL_SERVER_NO_ADD](m)

	def _handle_netobj_remove(self, m):
		self._NO._remove(PLd.handleNORemove(m))

	def _handle_netobj_updateinfo(self, m):
		"""
		This is an internal function which converts NetObject Update messges
		into Python data to be used by the NetObject callback function
		"""
		def datahandler(m):
			PLd.handleNOUpdate(m, self._NO._by_id)
					
		self._handler_add(PLd.PROTOCOL_SERVER_NO_UPDATE, datahandler)
		self.handlers[PLd.PROTOCOL_SERVER_NO_UPDATE](m)


	def _handle_service_dict(self, r):
		for k,v in r.iteritems():
			sn = self._strtbl.get(k, "Service not in stringtable (%d)" % k)
			if sn not in self._services:
				self._services[sn] = {}
			d = self._services[sn]
			d.update(v)
	def _handle_service_add(self, m):
		r = PLd.handleServiceAdd(m)
		self._handle_service_dict(r)
	def _handle_service_remove(self, m):
		r = PLd.handleServiceRemove(m)
	def _handle_service_update(self, m):
		r = PLd.handleServiceUpdate(m)
		self._handle_service_dict(r)

	def _process(self, timeout):
		"""
		This is an internal function that will process events as they
		occur on the socket
		"""

		inp = [self.s]

		if self.io.wantWrite():
			outp = [self.s]
		else:
			outp = []


		(i, o, e) = select.select(inp, outp, inp, timeout)

		if len(i) + len(o) + len(e) == 0:
			return False

		if len(e) > 0:
			""" Huston! We have a problem! """
			raise IOError("Select returned error")

		if len(o) > 0:
			if not self.io.doWrite():
				raise IOError("Error writing to socket")

		if len(i) > 0:
			if not self.io.doRead():
				raise IOError("Error reading from socket")

			m = self.io.readMsg()
			while m:
				cmd = m.cmd()
				if self.handlers.has_key(cmd):
					if self.handlers[cmd](m):
						break

				m = self.io.readMsg()
				
		return True

	def _do(self, func, cmd):
		class handler:
			def __init__(self, pld):
				self.pld = pld
				self.ret = None
			def __call__(self, msg):
				del self.pld.handlers[cmd]
				if func:
					self.ret = func(msg)
				return True

		h = handler(self)
		self._handler_add(cmd, h)

		while self.handlers.has_key(cmd):
			if not self._process(self.timeout):
				raise IOError("Operation timed out")

		return h.ret

	def close (self):
		"""
		Immediately disconnect from packetlogic system.

		After this method has been called this object becomes
		useless.

		It may be used to force a disconnect when the garbage
		collector can't be trusted to do a timely disconnect.
		"""
		class Closed(object):
			def __getattr__(self, n):
				raise IOError("Closed")
		self.s.close()
		self.__dict__.clear()
		self.__class__ = Closed
		
	def dyn_add(self, id, ip, subscriber=""):
		"""
		Add a dynamic netobject item

		Please note that netobject will not be verified with regard to
		performance and flexibility when adding to a uncommitted
		netobject.

			>>> rt.dyn_add(31682, '127.0.0.1')
			True

		@type id: C{int}
		@param id: the id number of the netobject to add the item to.

		@type ip: C{str}
		@param ip: the IP address to add.

		@type subscriber: C{str}
		@param subscriber: Subscriber name - or empty string to not create subscriber. Requires v12.0 firmware.

		@rtype: C{boolean}
		@returns: True if add was successful

		@raises PLDPermissionError: If the user lacks permission for requested action
		@raises PLDUnsupportedInFirmware: If subscriber argument is used and firmware doesn't support it.
		@raises ValueError: If an invalid IP address is used

		@attention: subscriber parameter requires v12.0 firmware or newer.
		"""
		if not verify_ip(ip):
			raise ValueError("Argument '%s' is not a valid IP address" % ip)

		PLd.dynAdd(self.io, id, ip, subscriber)
		if self._do(PLd.handleDynAdd, PLd.PROTOCOL_SERVER_DYNAMICNETOBJECT_ADD):
			return True
		return False


	def dyn_remove(self, id, ip):
		"""
		Remove a dynamic netobject item

			>>> rt.dyn_remove(31682, '127.0.0.1')
			True

		@type id: C{int}
		@param id: the id number of the netobject to remove the item from.

		@type ip: C{str}
		@param ip: the ip number to remove.

		@rtype: C{boolean}
		@returns: True if remove was successful

		@raises PLDPermissionError: If the user lacks permission for requested action
		"""		
		PLd.dynRemove(self.io, id, ip)
		if self._do(PLd.handleDynRemove, PLd.PROTOCOL_SERVER_DYNAMICNETOBJECT_REMOVE):
			return True
		return False

	class _ListCollector:
		def __init__(self, func):
			self.func = func
			self.lst = []

		def __call__(self, msg):
			self.func(msg, self.lst)

	def dyn_list(self):
		"""
		List all dynamic netobject items

			>>> rt.dyn_list()
			[(31589, '127.0.0.1'),
			 (31576, '192.168.1.1'),
			 (31589, '192.168.0.131'),
			 (31688, '192.168.0.131'),
			 (31577, '192.168.1.1'),
			 (31682, '127.0.0.1')]

		@rtype: C{list} of C{tuple}
		@return: a list of tuples containing the netobject id and the item.

		@raises PLDNotAvailable: If data is not ready yet, and you should retry this call later. (This is not an error.)
		@raises PLDPermissionError: If the user lacks permission for requested action
		"""

		PLd.dynList(self.io)
		return self._dyn_list_handle()

	def dyn_list_full(self):
		"""
		List all dynamic netobject items including subscriber names

			>>> rt.dyn_list()
			[(31589, '127.0.0.1', 'User 1'),
			 (31576, '192.168.1.1', 'User 123'),
			 (31589, '192.168.0.131', 'User 124'),
			 (31688, '192.168.0.131', 'User 124'),
			 (31577, '192.168.1.1', 'User 123'),
			 (31682, '127.0.0.1', 'User 1')]

		@rtype: C{list} of C{tuple}
		@return: a list of tuples containing the netobject id, the item and subscriber name.

		@raises PLDNotAvailable: If data is not ready yet, and you should retry this call later. (This is not an error.)
		@raises PLDPermissionError: If the user lacks permission for requested action
		@raises PLDUnsupportedInFirmware: If firmware doesn't support this function.

		@attention: This function requires v12.0 firmware or newer.
		"""

		PLd.dynList(self.io)
		return self._dyn_list_handle(full=True)

	def _dyn_list_handle(self, full=False):
		lc = self._ListCollector(PLd.handleDynEntries)
		slc = self._ListCollector(PLd.handleDynSubscriberEntries)
		self._handler_add(PLd.PROTOCOL_SERVER_DYNAMICNETOBJECT_LIST_ENTRIES,
				  lc)
		if full:
			self._handler_add(PLd.PROTOCOL_SERVER_DYNAMICNETOBJECT_LIST_SUBSCRIBER_ENTRIES,
					  slc)
		# run until we get DYNAMICNETOBJECT_LIST_END
		self._do(None, PLd.PROTOCOL_SERVER_DYNAMICNETOBJECT_LIST_END)
		del self.handlers[PLd.PROTOCOL_SERVER_DYNAMICNETOBJECT_LIST_ENTRIES]
		if full:
			del self.handlers[PLd.PROTOCOL_SERVER_DYNAMICNETOBJECT_LIST_SUBSCRIBER_ENTRIES]
			return map(lambda x,y:(x[0], x[1], y), lc.lst, slc.lst)
		else:
			return lc.lst

	def dyn_list_no(self, id):
		"""
		List dynamic netobject items matching the specified netobject id
		
		This command is NOT supported on system with PacketLogic version v11.359
		or earlier. Using it will raise a timeout exception after about a minute.

			>>> rt.dyn_list_no(31576)
			[(31576, '192.168.1.1'),
			 (31576, '192.168.0.131'),
			 (31576, '192.168.2.13')]

		@type id: C{int}
		@param id: the id number of the netobject to match against.

		@rtype: C{list} of C{tuple}
		@return: a list of tuples containing the netobject id and the item.

		@raises PLDNotAvailable: If data is not ready yet, and you should retry this call later. (This is not an error.)
		@raises PLDPermissionError: If the user lacks permission for requested action

		"""

		PLd.dynListNo(self.io, id)
		return self._dyn_list_handle()

	def dyn_list_mask(self, ip, mask):
		"""
		List all dynamic netobject items matching the specified ipaddress and netmask

		This command is NOT supported on system with PacketLogic version v11.359
		or earlier. Using it will raise a timeout exception after about a minute.

			>>> rt.dyn_list_mask('192.168.0.0', '255.255.252.0')
			[(31576, '192.168.1.1'),
			 (31589, '192.168.0.131'),
			 (31688, '192.168.0.131'),
			 (31577, '192.168.1.1'),
			 (31682, '192.168.2.33'),
			 (31697, '192.168.3.142')]

		@type ip: C{str}
		@param ip: the ip number to use for matching.

		@type mask: C{str}
		@param mask: the netmask to use for matching.

		@rtype: C{list} of C{tuple}
		@return: a list of tuples containing the netobject id and the item.

		@raises PLDNotAvailable: If data is not ready yet, and you should retry this call later. (This is not an error.)
		@raises PLDPermissionError: If the user lacks permission for requested action
		@raises PLDServerError: If either argument is of incorrect format.

		"""

		PLd.dynListMask(self.io, ip, mask)
		return self._dyn_list_handle()

	def vbs_reset(self, machine, soid, extra=0):
		"""
		Reset vbs accounting data for a vbs object.

		@type machine: C{str}
		@param machine: MachineID of the PacketLogic you want to query.

		@type soid: C{int}
		@param soid: ID of VBS object

		@type extra: C{int}
		@param extra: Extra key; netobject id in case of split
		              by netobject, or ip number in case of
		              split by local host (encoded as an
		              integer). See the return value of
		              L{vbs_query} for details.

		@raises PLDUnsupportedInFirmware: If firmware doesn't support this function.

		@attention: This function isn't supported in v12.0 and v12.1 firmwares.
		"""
		raise PLDUnsupportedInFirmware("VBS is not supported in this firmware version")

	def vbs_set(self, invalue, outvalue, machine, soid, extra=0):
		"""
		Set transfer to the specified value. Can only be used
		on object without sliding window (duration=1).

		@type invalue: C{int}
		@param invalue: The new value of inbound transfer (in KiB)

		@type outvalue: C{int}
		@param outvalue: The new value of outbound transfer (in KiB)

		@type machine: C{str}
		@param machine: MachineID of the PacketLogic you want to query.

		@type soid: C{int}
		@param soid: ID of VBS object

		@type extra: C{int}
		@param extra: Extra key; netobject id in case of split
		              by netobject, or ip number in case of
		              split by local host (encoded as an
		              integer). See the return value of
		              L{vbs_query} for details.

		@raises PLDUnsupportedInFirmware: If firmware doesn't support this function.

		@attention: This function does not exist in v12.0 firmware or newer.
		"""
		raise PLDUnsupportedInFirmware("VBS is not supported in this firmware version")

	def vbs_list(self):
		"""
		List available vbs objects

			>>> rt.vbs_list()
			[('00E081600B81', 24, 'VBS low', 1),
			 ('00E081600B81', 74, 'VBS medium', 1),
			 ('00E081600B81', 26, 'VBS max', 1)]

		@rtype: C{list} of C{tuple}
		@returns: A list of available VBS objects in a tuple (machineid,
		          id of vbs object, name, split type)
			  
			  split type is one of SPLIT_NONE, SPLIT_LOCALHOST,
			  SPLIT_HOST_NETOBJECT, SPLIT_SERVER_NETOBJECT,
			  SPLIT_CLIENT_NETOBJECT, SPLIT_CONNECTION

		@raises PLDPermissionError: If the user lacks permission for requested action
		@raises PLDUnsupportedInFirmware: If firmware doesn't support this function.

		@attention: This function isn't supported in v12.0 and v12.1 firmwares.
		"""
		raise PLDUnsupportedInFirmware("VBS is not supported in this firmware version")


	def vbs_query(self, machine, objid, extra=0):
		"""
		Query vbs data from a specific vbs object

			>>> rt.vbs_query('0010F3067C3A', 24)
			[{'bidirectional': 0L,
			  'bidirectional speed': 0,
			  'extra': -183572567,
			  'id': 24,
			  'inbound': 0L,
			  'inbound speed': 0,
			  'machineid': '0010F3067C3A',
			  'name': 'VBS Grund',
			  'outbound': 0L,
			  'outbound speed': 0,
			  'splitname': '169.231.14.245'}]

		@type machine: C{str}
		@param machine: MachineID of the PacketLogic you want to query.

		@type objid: C{int}
		@param objid: ID of VBS shapingobject

		@type extra: C{int}
		@param extra: Extra key; netobject id in case of split
		              by netobject, or ip number in case of
		              split by local host (encoded as an
		              integer). If argument isn't given ALL
		              available split keys are listed.

		@rtype: C{list} of C{dict}

		@returns: A list of dictionaries, where each
		          dictionary holds VBS data with the following
		          contents: 
		           - 'machineid', 'id', 'name'
		                identify the shaping object.
		           - 'inbound/outbound/bidirectional'
		                accumulated volume of transferred data
		           - 'inbound/outbound/bidirectional speed'
		                holds the current speed
		           - 'splitname'
		                text representation of the split instance of the
		                shaping object (for example for split by
		                local host, the local host IP address as a
		                string).
		           - 'extra' 
		                numerical representation of the split
		                instance of the shaping object (for
		                example for split by local host; an
		                integer representation of the IP
		                address, for split by netobject; the
		                id number of the netobject). This is
		                the number that would be passed as
		                'extra' argument to L{vbs_reset} or
		                L{vbs_query}.

		@raises PLDPermissionError: If the user lacks permission for requested action

		@attention: This function isn't supported in v12.0 and
		            v12.1 firmwares.
		@attention: In v12.2 firmware or newer the "machineid"
		            argument is ignored.
		@attention: In v12.2 firmware or newer bidirectional
		            VBS isn't supported, therefore the
		            'bidirectional' value returned is always 0.
		"""
		PLd.vbsQuery(self.io, machine, objid, extra)
		lc = self._ListCollector(PLd.handleVbsQuery)
		self._handler_add(PLd.PROTOCOL_SERVER_VBS_REPLY, lc)
		# run until we get VBS_REPLY_END
		self._do(None, PLd.PROTOCOL_SERVER_VBS_REPLY_END)
		del self.handlers[PLd.PROTOCOL_SERVER_VBS_REPLY]
		return lc.lst


	def add_stats_callback(self, func):
		"""
		Add a callback to call with system diagnostics data
		while updating.
		
		@type func: C{callable}
		@param func: function to call. The function is called
		             with a huge dict containing all system
		             diagnostics data.

		"""

		self.stats_callbacks.append(func)
		if len(self.stats_callbacks) > 1:
			# No need to add at server again
			return

		# Tell server we want stats..
		PLd.statsShow(self.io)

	def add_shapingcnt_callback(self, func):
		"""Add a callback for byte counters.

		The signature of the callback should be:

		C{cb(idx, oid, sk, flags, _in, _out)}
		
		idx is counter index in PL (i.e a number between 0 and MAX_COUNTERS)
		oidx is shaping object id
		sk is split key (i.e netobject id on split-by-netobject, ip on
		split-by-localhost, subscriber name on split-by-subscriber, and 0 on no
		split)
		flags is flags bitmask, if (1<<13) is set it means that this is the last
		update for this counter, and it will be reset to 0.
		_in and _out are bytecounters.

		@raises PLDUnsupportedInFirmware: If firmware doesn't support this function.

		@attention: This function requires v12.2 firmware or newer.
		"""
		raise PLDUnsupportedInFirmware("add_shapingcnt_callback is not supported in this firmware version")

	def shapingcnt_request_all(self):
		"""
		Requests all shaping counters, regardless if they have
		changed or not.
		
		@attention: This function requires v12.2 firmware or newer.
		"""
		raise PLDUnsupportedInFirmware("shapingcnt_request_all is not supported in this firmware version")


	def add_update_done_callback(self, func):
		"""
		Add a callback to call after each update is done.

		@type func: C{callable}
		@param func: function to call. The function is called
		             without any arguments

		"""
		self.update_done_callbacks.append(func)

	def _request_debug_stats(self):
		PLd.statsShowDebug(self.io)

	def add_channels_callback(self, func):
		"""
		Add a callback to call with channel statistics data
		while updating.
		
		@type func: C{callable}
		@param func: function to call. The function is called
		             with a dict with the channel ids as key
		             and a inner dict as value.

		"""

		self.chans_callbacks.append(func)
		if len(self.chans_callbacks) > 1:
			# No need to add at server again
			return

		# Tell server we want stats..
		PLd.channelsShow(self.io)

	def add_netobj_callback(self, func, under=None):
		"""
		Add a callback to call with realtime NetObject
		data while updating.
		
		@type func: C{callable}
		@param func: function to call. The function is called
		             with one argument, an L{RealtimeNetObjectTree}
			     containing the list of L{RealtimeNetObject}s data,
			     which is fetched with that objects methods.

		@type under: C{str}
		@param under: Request netobject data under a specified
		              path. In v12.0 and newer only toplevel
		              objects are included by default, and
		              this in required to get data for a
		              specific path. Observe that only the
		              objects directly under the specified
		              path is reported, no recursive
		              listing. When 'under' is used the
		              callback is called with a list of
		              L{RealtimeNetObject}s directly, no
		              L{RealtimeNetObjectTree} intermediate
		              layer.

		@raises PLDUnsupportedInFirmware: If under argument is used and firmware doesn't support it.

		@attention: under parameter requires v12.1 firmware or newer.
		"""
		if under is not None:
			if under[-1] != '/':
				under = under + "/"
			self.interesting_netobjects.append(under)

		self.no_callbacks.append((func, under))
		# Add a handler to NO_ADD messages (not done before, to prevent no-msg:s sent on commit to allocate memory, when not needed.)
		self._handler_add(PLd.PROTOCOL_SERVER_NO_ADD, self._handle_netobj_add)
		self._handler_add(PLd.PROTOCOL_SERVER_NO_REMOVE, self._handle_netobj_remove)

		if len(self.no_callbacks) > 1:
			# No need to add at server again
			return

		# Tell server we want stats.. (For NetObjects we must also request an update to actually get any data)
		PLd.netobjectsShow(self.io)
		PLd.requestUpdate(self.io)

	def add_service_callback(self, func):
		"""
		Add a callback to call with realtime service
		data while updating.
		
		@type func: C{callable}
		@param func: function to call. The function is called
		             with one argument, a C{dict} of services.
			     Key is servicename, value is a C{dict} of its properties.

		@raises PLDUnsupportedInFirmware: If firmware doesn't support this function.

		@attention: This function requires v12.1 firmware or newer.
		"""
		self.service_callbacks.append(func)

		if len(self.service_callbacks) > 1:
			# No need to add at server again
			return

		# Tell server we want stats.. (For NetObjects we must also request an update to actually get any data)
		PLd.servicesShow(self.io)
		PLd.requestUpdate(self.io)


	def get_channels_data(self, maxage=5):
		"""
		Get channel statistics data once.
		
		(Should not be used simultaneously to
		add_channels_callback and update_forever)

		@type maxage: C{int}
		@param maxage: Max age in seconds of cached information.
		               If longer time has passed since last call,
			       new data will be requested from PacketLogic.
			     
		@raises PLDPermissionError: If the user lacks permission.
		
		@rtype: C{dict}
		@returns: Dict containing all channels. Channel id is used
		          as key and the value is another dict containing
		          values.
		
		"""
		if self._cache_channels is None or time.time() >= self._cache_channels_ts + maxage:
			self.add_channels_callback(self._update_once_callback)
			self._cache_channels = self._update_once()
			self._cache_channels_ts = time.time()
		
		return self._cache_channels

	def get_stats_data(self, maxage=5):
		"""
		Get system diagnostics data once.

		(Should not be used simultaneously to
		add_stats_callback and update_forever)
		
		@type maxage: C{int}
		@param maxage: Max age in seconds of cached information.
		               If longer time has passed since last call,
			       new data will be requested from PacketLogic.
			     
		@raises PLDPermissionError: If the user lacks permission.
		
		@rtype: C{dict}
		@returns: Dict containing all system diagnostics data in
		          the form of nested dicts, the toplevel dict
		          contains contains the different system diagnostics
		          zones.
		
		"""
		if self._cache_stats is None or time.time() >= self._cache_stats_ts + maxage:
			self.add_stats_callback(self._update_once_callback)
			self._cache_stats = self._update_once()
			self._cache_stats_ts = time.time()
		
		return self._cache_stats
		
	def get_netobj_data(self, maxage=5):
		"""
		Get Realtime NetObject data once.

		(Should not be used simultaneously to
		add_netobj_callback and update_forever)
		
		@type maxage: C{int}
		@param maxage: Max age in seconds of cached information.
		               If longer time has passed since last call,
			       new data will be requested from PacketLogic.
			     
		@raises PLDPermissionError: If the user lacks permission.
		
		@rtype: L{RealtimeNetObjectTree}
		@returns: An object that is iterable to fetch all NetObjects,
			otherwise it has methods to fetch specific data.
		
		"""
		if time.time() >= self._netobj_ts + maxage:
			self.add_netobj_callback(self._update_once_callback)
			self._update_once()
			self._NO._link_parents()
			self._netobj_ts = time.time()
		
		return self._NO

	def get_services_data(self, maxage=5):
		"""
		Get Realtime Services data once.

		(Should not be used simultaneously to
		add_service_callback and update_forever)
		
		@type maxage: C{int}
		@param maxage: Max age in seconds of cached information.
		               If longer time has passed since last call,
			       new data will be requested from PacketLogic.
			     
		@raises PLDPermissionError: If the user lacks permission.
		@raises PLDUnsupportedInFirmware: If firmware doesn't support this function.

		@attention: This function requires v12.1 firmware or newer.
		"""
		if time.time() >= self._service_ts + maxage:
			self.add_service_callback(self._update_once_callback)
			self._update_once()
			self._service_ts = time.time()
		
		return self._services
		
	def _update_once(self):
		""" Wait for a stats/channel callback and return."""
		last = time.time()
		self.updating = True
		self._update_once_data = None
		PLd.requestUpdate(self.io)
		timeout = self.timeout
		if timeout is None:
			timeout = sys.maxint
		while self.updating:
			if time.time() > last + timeout:
				raise IOError("Operation timed out")
			if self._process(5):
				last = time.time()
				
		return self._update_once_data

	def _update_once_callback(self, data):
		""" Callback for the _update_once function."""
		
		self._update_once_data = data
		self.stop_updating()
		
	def update_forever(self, interval=5):
		"""
		Start updating.

		@type interval: C{int}
		@param interval: Delay between updates.

		@raises PLDPermissionError: If the user lacks permission for any of the registerd callbacks

		"""

		next = time.time()
		last = time.time()
		self.updating = True
		timeout = self.timeout
		if timeout is None:
			timeout = sys.maxint
		while self.updating:
			now = time.time()
			if now > last + timeout:
				raise IOError("Operation timed out")
			if now > next:
				PLd.requestUpdate(self.io)
				next = now + interval
			if self._process(next - now):
				last = time.time()

	def stop_updating(self):
		"""
		Stop the updating and clear callback functions.

		"""
		self.updating = False
		self.stats_callbacks = []
		self.chans_callbacks = []
		self.no_callbacks = []
		self.service_callbacks = []
		PLd.statsHide(self.io)
		PLd.channelsHide(self.io)
		PLd.netobjectsHide(self.io)
		PLd.servicesHide(self.io)
		# Remove the add-handler if we have one, no need to waste time running it when until we need the data again.
		self._handler_remove(PLd.PROTOCOL_SERVER_NO_ADD)

	def fwlog_clear(self):
		"""
		Clear the firewall log.

		@returns: Nothing

		@raises PLDPermissionError: If the user lacks permission for requested action
		"""
		PLd.fwlog_clear(self.io)
		self._flush()

	def _fwlog_query(self, qtype, start_or_id, end, clear=False, no_reverse=False):
		"""
		Query the firewall log for a list of entries

		@type qtype: C{int}
		@param qtype: Query type
			* 0 List of all entries matching a single rules
			* 1 List of all entries between two points in time
			* 2 List of all M entries offset at O. Suitable for access one "page" at a time.
		
		@type clear: C{bool}
		@param clear: If True, log is cleared after values are retrieved. This clears the whole log and not the selected portion. Not supported in v12.0 firmwares.

		@type no_reverse: C{bool}
		@param no_reverse: If True, the offsets are handled from oldest entry, not newest entry. i.e fwlog_query_offsets(0,10,no_reverse=True) will list 10 oldest entries, not 10 newest. Not supported in v12.0 firmwares.

		@returns: 2-tuple with firewall log information and matching entries.
		@rtype: C{tuple} of C{list} and C{dict}

		@raises PLDPermissionError: If the user lacks permission for requested action
		@raises PLDUnsupportedInFirmware: If clear or no_reverse parameter is used on v12.0 firmware.

		@attention: clear parameter is not supported on v12.0 firmwares.
		@attention: no_reverse parameter is not supported on v12.0 firmwares.
		"""

		class LC_entry:
			def __init__(self, func):
				self.func = func
				self.lst = []
			def __call__(self, msg):
				self.lst.append(self.func(msg))
		class LC_info:
			def __init__(self, func):
				self.func = func
			def __call__(self, msg):
				self.info = self.func(msg)
				
				
		PLd.fwlog_query(self.io, qtype, start_or_id, end, clear, no_reverse)

		lc1 = LC_info(PLd.handleFWEntryStart)
		self._handler_add(PLd.PROTOCOL_SERVER_FWLOG_ENTRY_START, lc1)
		
		lc2 = LC_entry(PLd.handleFWEntry)
		self._handler_add(PLd.PROTOCOL_SERVER_FWLOG_ENTRY, lc2)

		# run until we get _END
		self._do(None, PLd.PROTOCOL_SERVER_FWLOG_ENTRY_END)

		del self.handlers[PLd.PROTOCOL_SERVER_FWLOG_ENTRY_START]
		del self.handlers[PLd.PROTOCOL_SERVER_FWLOG_ENTRY]

		return (lc1.info, lc2.lst)
		

	def fwlog_query_rule(self, id):
		"""
		Query the firewall log for a list of entries matching a specific firewall rule

			>>> info, results = rt.fwlog_query_rule(58)
			>>> print info
			{'logentries': 14054L}
			>>> print results
			[{'client': '10.0.0.2',
			  'clienthostname': '',
			  'clientport': 2985,
			  'properties': '',
			  'protocol': 6,
			  'rule': 'Monitor ',
			  'server': '10.0.1.3',
			  'serverhostname': '',
			  'serverport': 1512,
			  'service': 'Skype',
			  'time': 1138204680,
			  'type': 0},
			 {'client': '1.2.3.4',
			  'clienthostname': '',
			  'clientport': 3247,
			  'properties': '',
			  'protocol': 6,
			  'rule': 'Monitor ',
			  'server': '10.3.2.1',
			  'serverhostname': '',
			  'serverport': 17561,
			  'service': 'Undetermined',
			  'time': 1138204680,
			  'type': 0},
			 {'client': '3.2.1.2',
			  'clienthostname': '',
			  'clientport': 52263,
			  'properties': 'Serverinfo= AAMAILE.int.aalogistik.com Microsoft ESMTP MAIL Service, Version: 6.0.3790.1830 ready at  Wed, 25 Jan 2006 16:57:54 +0100 \\n',
			  'protocol': 6,
			  'rule': 'Monitor ',
			  'server': '1.2.4.3',
			  'serverhostname': '',
			  'serverport': 25,
			  'service': 'Undetermined',
			  'time': 1138204680,
			  'type': 0}]

		@type id: C{int}
		@param id: Rule id
		
		@returns: 2-tuple with firewall log information and matching entries.
		@rtype: C{tuple} of C{dict} and C{list}

		@raises PLDPermissionError: If the user lacks permission for requested action
		"""

		return self._fwlog_query(PLd.PROTOCOL_FWRULE_QTYPE_ID, id, 0)

	

	def fwlog_query_offsets(self, first_nr, last_nr, clear=False, no_reverse=False):
		"""
		Query the firewall log for a list of entries between two offsets

			>>> info, results = rt.fwlog_query_offsets (0, 50)
			>>> print info
			{'logentries': 14054L}
			>>> print len(results)
			50
			>>> for i in results[0:3]:
			....:     print i
			....:
			{'clientport': 39682, 'protocol': 6, 'service': 'Lunarstorm live', 'clienthostname': '', 'serverhostname': '', 'rule': 'Monitor ', 'server': '1.2.3.4', 'client': '5.6.7.8', 'serverport': 8080, 'time': 1138204688, 'type': 0, 'properties': ''}
			{'clientport': 55172, 'protocol': 6, 'service': 'HTTP', 'clienthostname': '', 'serverhostname': 'www.google.com', 'rule': 'Log HTTP', 'server': '209.85.129.104', 'client': '192.168.9.174', 'serverport': 80, 'time': 1250775399L, 'type': 0, 'properties': 'File length=0\\nServer Hostname=www.google.com\\nURL=http://www.google.com/search?q=packetlogic&ie=utf-8&oe=utf-8&aq=t\\nRequest method=GET\\nDirection=Inbound\\nFilename=/search?q=packetlogic&ie=utf-8&oe=utf-8&aq=t\\nUser-Agent=Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.9.0.12) Gecko/2009072221\\nTransfer-Encoding=None\\nUpload file length=0\\n'}
			{'clientport': 3252, 'protocol': 6, 'service': 'eDonkey transfer', 'clienthostname': '', 'serverhostname': '', 'rule': 'Monitor ', 'server': '10.10.11.11', 'client': '40.50.60.70', 'serverport': 7804, 'time': 1138204688, 'type': 0, 'properties': ''}


		Grab all entries in the firewall log and clear once we get the last batch
		
			>>> l = []
			>>> numentries = rt.fwlog_query_offsets(0,0)[0]['logentries']
			>>> for x in range(0, numentries, 500):
			....:	l += rt.fwlog_query_offsets(x, min(x+500, numentries), no_reverse=True)[1]
			....:
			>>> l += rt.fwlog_query_offsets(numentries, -1, clear=True, no_reverse=True)[1] # grab entries logged while quering



		@type first_nr: C{str}
		@param first_nr: Start the list at this offset. 0 means the start of the firewall log. 
		
		@type last_nr: C{str}
		@param last_nr: End the list at this offset.
		
		@type clear: C{bool}
		@param clear: If True, log is cleared after values are retrieved. This clears the whole log and not the selected portion. Not supported in v12.0 firmwares.

		@type no_reverse: C{bool}
		@param no_reverse: If True, the offsets are handled from oldest entry, not newest entry. i.e fwlog_query_offsets(0,10,no_reverse=True) will list 10 oldest entries, not 10 newest. Not supported in v12.0 firmwares.

		@returns: 2-tuple with firewall log information and matching entries. last_nr - first_nr entries will be returned.
		@rtype: C{tuple} of C{list} and C{dict}

		@raises PLDPermissionError: If the user lacks permission for requested action
		@raises PLDUnsupportedInFirmware: If clear or no_reverse parameter is used on v12.0 firmware.

		@attention: clear parameter is not supported on v12.0 firmwares.
		@attention: no_reverse parameter is not supported on v12.0 firmwares.
		"""

		return self._fwlog_query(PLd.PROTOCOL_FWRULE_QTYPE_OFFSETS, first_nr, last_nr, clear, no_reverse)

	def fwlog_query_time_range(self, timefr, timeto):
		"""
		Query the firewall log for a list between two points in time

			>>> info,results = rt.fwlog_query_time_range ("2006-01-01 00:00", "2006-02-02 00:00")
			>>> print results[0]
			{'clientport': 38742, 'protocol': 6, 'service': 'eDonkey transfer', 'clienthostname': '', 'serverhostname': '', 'rule': 'Monitor ', 'server': '3.3.3.3', 'client': '1.1.1.1', 'serverport': 4662, 'time': 1138204658, 'type': 0, 'properties': ''}

		@type timefr: C{str}
		@param timefr: Start the list at this date, the date has to be in I{YYYY-MM-DD HH:MM} format
		
		@type timeto: C{str}
		@param timeto: End the list at this date, the date has to be in I{YYYY-MM-DD HH:MM} format
			
		@returns: 2-tuple with firewall log information and matching entries.
		@rtype: C{tuple} of C{list} and C{dict}

		@raises PLDPermissionError: If the user lacks permission for requested action
		"""

		timefr = time.mktime(time.strptime(timefr, "%Y-%m-%d %H:%M"))
		timeto = time.mktime(time.strptime(timeto, "%Y-%m-%d %H:%M"))

		""" We have the time in UTC but we want to send it in localtime """
		timefr -= time.altzone
		timeto -= time.altzone

		return self._fwlog_query(PLd.PROTOCOL_FWRULE_QTYPE_FROMTO, long(timefr), long(timeto))

	def vcrc(self, criteria):
		"""
		Run virtual connection ruleset checker.

			>>> rt.vcrc({'serverip':'192.168.9.1'})


			>>> rt.vcrc({'serverip': '192.168.9.1',
				    'service': 'HTTP',
				    'vlan': 1234})

			>>> rt.vcrc({'serverip':'217.73.103.54'})
			{'fwrules': [(60L, 2031L), (63L, 2031L)], 'shapingrules': [(65L, 2031L),
			(66L, 1903L)], 'monitor_iface': 0L, 'fwrule': 63, 'ruleset_state': 3,
			'test_id': 1L, 'rewrite_obj': 0L}
		
		@type criteria: C{dict}
		@param criteria: Dict with any of the following keys:

		 * vlan
		 * mpls
		 * protocol
		 * clientip
		 * serverip
		 * clientport
		 * serverport
		 * service
		 * xfb
		 * properties

		@returns: dict with information, including lists for shaping and firewall rules with a tuple per matching rule. The first element in each tuple is the rule id,
		and the second element is a bitmask which describes what criteria for that rule are mutable. Mutable in this context means that there exist values for that
		criteria which will make the rule not match.
		The bitmask has the following bits defined:
		PL_RULESET_VIRTUAL_COND_VLAN = 0
		PL_RULESET_VIRTUAL_COND_MPLS = 1
		PL_RULESET_VIRTUAL_COND_PROTOCOL = 2
		PL_RULESET_VIRTUAL_COND_CLIENTIP = 3
		PL_RULESET_VIRTUAL_COND_SERVERIP = 4
		PL_RULESET_VIRTUAL_COND_CLIENTPORT = 5
		PL_RULESET_VIRTUAL_COND_SERVERPORT = 6
		PL_RULESET_VIRTUAL_COND_SERVICE = 7
		PL_RULESET_VIRTUAL_COND_ASPATH = 8
		PL_RULESET_VIRTUAL_COND_XFB = 9
		PL_RULESET_VIRTUAL_COND_PROPERTIES = 10
		The ruleset_state returned in the dict signifies the final firewall state for the connection. It is one of the following:
		PL_RULESET_ACTION_UNKNOWN = 0
		PL_RULESET_ACTION_ACCEPT = 1
		PL_RULESET_ACTION_REJECT = 2
		PL_RULESET_ACTION_DROP = 3
		PL_RULESET_ACTION_REWRITE = 4
		PL_RULESET_ACTION_DIVERT = 5
		"""

		PLd.vcrcTest(self.io, 1, criteria)
		return self._do(PLd.handleVcrcTest, PLd.PROTOCOL_SERVER_VCRC_TEST_RESULT)
		

	def dump_bgptable(self):
		"""
		Dump the bgp table.

		This function is just for debugging purposes. It will return
		the BGP table in packetlogicd.

		@returns: tuple with all prefix and paths.
		@rtype: C{tuple} of C{list}

		@raises PLDPermissionError: If the user lacks permission for requested action
		@raises PLDUnsupportedFeatureError: If the 'bgp' feature isn't enabled.
		"""
		
		self._require_features('bgp')
		
		class LC:
			def __init__(self, func):
				self.func = func
				self.lst = []
				
			def __call__(self, msg):
				self.lst.append(self.func(msg))


		PLd.bgpDump(self.io)

		lc1 = LC(PLd.handleBgpEntry)
		self._handler_add(PLd.PROTOCOL_SERVER_BGP_DUMP_ENTRY, lc1)

		lc2 = LC(PLd.handleBgpPath)
		self._handler_add(PLd.PROTOCOL_SERVER_BGP_DUMP_PATH, lc2)

		# run until we get _DONE
		self._do(None, PLd.PROTOCOL_SERVER_BGP_DUMP_DONE)

		del self.handlers[PLd.PROTOCOL_SERVER_BGP_DUMP_ENTRY]
		del self.handlers[PLd.PROTOCOL_SERVER_BGP_DUMP_PATH]

		return (lc1.lst, lc2.lst)

	def dump_drdlstats(self):
		"""
		Dump drdl stats

		This function is just for debugging purposes. You probably
		don't know what the data means anyway :)

		@returns: the stats
		@rtype: C{list}

		@raises PLDPermissionError: If the user lacks permission for requested action
		"""

		PLd.drdlStatsDump(self.io)

		lst = []

		def h(msg):
			PLd.handleDrdlStatsDump(msg, lst)
		self._handler_add(PLd.PROTOCOL_SERVER_DRDL_STATS_ENTRIES, h)

		# run until we get _DONE
		self._do(None, PLd.PROTOCOL_SERVER_DRDL_STATS_END)

		del self.handlers[PLd.PROTOCOL_SERVER_DRDL_STATS_ENTRIES]

		return lst

	def drdl_revision(self):
		"""
		Retrieve revision string for currently loaded drdl ARM.

		@returns: Revision string.
		@rtype: C{str}

		@raises PLDPermissionError: If the user lacks permission for requested action
		@raises PLDUnsupportedInFirmware: If firmware doesn't support this function.

		@attention: This function requires v12.1 firmware or newer.
		"""
		PLd.drdlRevision(self.io)
		return self._do(PLd.handleDrdlRevision, PLd.PROTOCOL_SERVER_DRDL_REVISION)

	def reload_drdl(self):
		"""
		Reload drdl signatures. For internal use only.

		@raises PLDUnsupportedInFirmware: If firmware doesn't support this function.

		@attention: This function requires v12.1 firmware or newer.
		"""

		PLd.drdlReload(self.io)
		self._flush()

	def inject_packet(self, iface, pkt):
		"""
		Inject a packet.

		This function is just for testing purposes.

		@raises PLDPermissionError: If the user lacks permission for requested action

		@raises ValueError: If the size of a packet exceeds 4096 bytes.
		"""

		PLd.injectPacket(self.io, iface, pkt)
		self._flush()

	def force_statistics_write(self):
		"""
		Force Statistics write to the disk.

		@raises PLDUnsupportedInFirmware: If firmware doesn't support this function.

		@attention: This function requires v12.1 firmware or newer.
		"""
		PLd.force_statistics_write(self.io)
		self._flush()


class StatsZone(PLd.ClientStatsZone):
	"""Class for registering system diagnostics zone.

			>>> zone = pld.StatsZone('TestStats')
			>>> zone.add_size('Entries')
			>>> zone.add_timespan('Uptime')
			>>> zone.register(r)
			>>> while True:
			...     zone['Entries'] = 1337
			...     zone['Uptime'] = time.time() - starttime
			...     zone.send(r)
			...     time.sleep(3)

		@raises KeyError: If an invalid key is specified using [ ] index.
	"""
	

	def __init__(self, name, idx=0):
		"""

		@type name: C{str}
		@param name: Name of the system diagnostics zone to registered
		
		@type idx: C{int}
		@param idx: Zone index, used if multiple zones are to be registered. Only available in v12.1 firmware and newer.

		@raises PLDUnsupportedInFirmware: If idx is used without firmware support.

		@attention: Registering multiple zones requires v12.1 firmware or newer.
		"""
		PLd.ClientStatsZone.__init__(self, name, idx)
		self.values = {}
		
	def register(self, pld):
		"""
		Register zone to PLd connection.
		"""
		PLd.ClientStatsZone.register(self, pld.io)
		pld._flush()

	def send(self, pld):
		"""
		
		"""
		PLd.ClientStatsZone.send(self, pld.io)
		pld._flush()
		
	def add(self, name, typ):
		"""
		@raises RuntimeError: if this StatsZone has been registered already
		
		@raises ValueError: if a value with the given name is added already.
		"""
		PLd.ClientStatsZone.add(self, name, typ)
		self.values[name] = None

	def add_interval(self, name):
		self.add(name, PLd.STATS_TYPE_INTERVAL)

	def add_moment(self, name):
		self.add(name, PLd.STATS_TYPE_MOMENT)

	def add_size(self, name):
		self.add(name, PLd.STATS_TYPE_SIZE)

	def add_speed(self, name):
		self.add(name, PLd.STATS_TYPE_SPEED)

	def add_avg(self, name):
		self.add(name, PLd.STATS_TYPE_AVG)

	def add_timespan(self, name):
		self.add(name, PLd.STATS_TYPE_TIMESPAN)

	def add_percent(self, name):
		self.add(name, PLd.STATS_TYPE_PERCENT)

	def add_time(self, name):
		self.add(name, PLd.STATS_TYPE_TIME)

	def __setitem__(self, name, value):
		if self.values.has_key(name):
			self.set(name, value)
			self.values[name] = value
			return
		raise KeyError, name
	
	def __getitem__(self, name):
		if self.values.has_key(name):
			return self.values[name]
		raise KeyError, name

class RealtimeNetObject(object):
	"""
	Object containing Realtime NetObject data.

	@type id: C{int}
	@ivar id: The internal id.
	
	@type parent: L{RealtimeNetObject} or C{None}
	@ivar parent: References the parent NetObject, if any.
	
	@type name: C{str}
	@ivar name: Name of NetObject.
	
	@type fullpath: C{str}
	@ivar fullpath: Path including Name.
	
	@type speed: C{tuple} of two C{int}s
	@ivar speed: Speed in byte/s in/out.
	
	@type forwarded: C{tuple} of two C{int}s
	@ivar forwarded: Forwarded bytes in/out. Obsolete in v12.0 and newer, will always be 0 in these versions.
	
	@type connections: C{tuple} of two C{int}s
	@ivar connections: Connections in/out.
	
	@type cps: C{tuple} of two C{int}s
	@ivar cps: Connections per second in/out.
	
	"""
	#__slots__ = ('id', 'parent', 'path', 'name', 'speed', 'forwarded', 'connections', 'cps')
	__slots__ = ('id', 'parent', 'fullpath', 'name', 'speed', 'forwarded', 'connections', 'cps')
	
	def __init__(self):
		"""
		This class is instantiated internally, no need to instantiate it manually.
		"""
		self.id = 0L
		self.parent = 0L
		self.name = ''
		self.speed = (0L,0L)
		self.forwarded = (0L,0L)
		self.connections = (0L,0L)
		self.cps = (0L, 0L)

	def __repr__(self):
		return "<%s '%s' %d>" % (self.__class__.__name__, self.name, self.id)

	fullpath = property(lambda self: "%s/%s" % (self.path, self.name), doc="Full path of this object including the object's name.")
	def _getpath(self):
		if not self.parent:
			return '/NetObjects'
		else:
			return self.parent.fullpath
	
	path = property(_getpath, doc="Path which this object resides under.")

class RealtimeNetObjectTree(object):
	"""
	NetObject data accessed through the NetObject callback function

		>>> import packetlogic2
		>>> pl = packetlogic2.connect('192.168.1.25', 'admin', 'pldemo00')
		>>> rt = pl.Realtime()
		>>> count = 1
		>>> def f(data):
		...	global count
		...	
		...	if count > 100:
		...		rt.stop_updating()
		...	if data:
		...		for m in data.list('/NetObjects/DSL'):
		...			print "name: %s" % m.fullpath
		...			print "speed: ", m.speed
		...			print "connections: ", m.connections
		...		m = data.find('/NetObjects/DSL/Lab')
		...		if m is not None:
		...			print "name: %s" % m.fullpath
		...			print "speed: ", m.speed
		...			print "connections: ", m.connections
		...	count +=1
		...
		>>> rt.add_netobj_callback(f)
		>>> rt.update_forever()			

	"""
	__slots__ = ('_cache', '_by_id')
	
	def __init__(self):
		self._clear()
		
	def _clear(self):
		self._cache = []
		self._by_id = {}

	def __len__(self):
		return len(self._cache)

	def _append(self, netobj):
		self._cache.append(netobj)
		self._by_id[netobj.id] = netobj
		
	def _extend(self, netobj_list):
		for netobj in netobj_list:
			self._append(netobj)

	def _removeobj(self, no):
		for x in self._cache[:]:
			if x.parent is no:
				self._removeobj(x)
		self._cache.remove(no)
		del self._by_id[no.id]

	def _remove(self, noid):
		no = self._by_id.get(noid)
		if no is None:
			print "Remove of nonexistant object?"
			return
		assert no.id == noid
		self._removeobj(no)
		
	def find(self, path):
		"""
		Find a single Realtime NetObject with the given full path.
		
		@returns: The specified NetObject data, or None if not found.
		@rtype: L{RealtimeNetObject}
		"""
		for obj in self._cache:
			if obj.fullpath == path:
				return obj
		return None

	#def getlist(self, path='/NetObjects'):
	#	if self._cache and isinstance(self._cache[0].parent, (int, long)):
	#		self._link_parents()
	#	return [obj for obj in self._cache if obj.getfullpath.startswith(path + "/") or obj.getfullpath == path]
			
	def list(self, path='/NetObjects', recursive=True):
		"""
		Return a list of Realtime NetObject data, matching the
		given path, and optional recursive flag.

		@type path: C{str}
		@param path: Path where to start looking. If the path
			ends with a slash (in recursive mode), it will
			include the children of that NetObject (if any)
			but not the NetObject itself.
			The path should begin with '/NetObjects'.

		@type recursive: C{bool}
		@param recursive: If True any level of children NetObjects will
			also be returned. Otherwise only the NetObjects
			that is direct children to the given full path
			is returned.
			(Do not end path with / if recursive is True).
		
		@returns: List of matching Realtime NetObjects.
		@rtype: C{list} of L{RealtimeNetObject}s
		"""
		if recursive:
			if path.endswith('/'):
				return [obj for obj in self._cache if obj.fullpath.startswith(path)]
			else:
				return [obj for obj in self._cache if obj.fullpath.startswith(path + "/") or obj.fullpath == path]
		else:
			return [obj for obj in self._cache if obj.path == path]

	def __iter__(self):
		return self._cache.__iter__()
	
	def _link_parent_one(self, netobj):
		if netobj.parent == 0:
			netobj.parent = None
		elif isinstance(netobj.parent, RealtimeNetObject):
			pass
		else:
			netobj.parent = self._by_id.get(netobj.parent, netobj.parent)

	def _link_parents(self):
		"""
		Replace property parent with that object, not the id of it.
		Root NetObjects will have parent=None.
		"""
		for netobj in self._cache:
			self._link_parent_one(netobj)
