"""
This module contains functionality for managing backups on the
PacketLogic system.

Do NOT import this module directly. The L{PLConnection
<packetlogic2.PLConnection>} object should be used, e.g:

    >>> import packetlogic2
    >>> pl = packetlogic2.connect("192.168.1.25", "admin", "pldemo00")
    >>> b = pl.Backup()

@var __revision__: Revision identifier.
"""

__revision__ = '%K%'

from resource import Resource
from thisversion import PLDB

import plutil

class Backup(Resource):
	"""
	A connection to a PacketLogic's "Backup" resource.

	The methods of this object allows listing, taking and
	restoring backups on the PacketLogic system.

	@group Backup: list,take,restore
	"""
	def _resource(self):
		return PLDB.RESOURCE_BACKUPS

	def list(self):
		"""
		List backups stored on server.

			>>> b.list()
			[{'cdate': '', 'path': '20090520-05.14.plb', 'size': 353714L},
			 {'cdate': '', 'path': '20090731-22.52.plb', 'size': 272156L},
			 {'cdate': '', 'path': '20090831-09.56.plb', 'size': 177219L},
			 {'cdate': '', 'path': '20090916-16.24.plb', 'size': 596406L}]

		@rtype: C{list} of C{dict}
		@returns: a list of backups

		"""

		return plutil._do(self, self.session.backup_list(), plutil._ListCB(self, self.session.backup_list_handle, None, False))

	def take(self):
		"""
		Take a new backup.

		Creates a new backup of current database contents. The
		backup will be given the current date and time as
		name.

			>>> b.take()
			'20091026-08.15.plb'

		@rtype: C{str}
		@returns: Name of new backup
		"""
		return plutil._do(self, self.session.backup_dump(), plutil._AddCB(self, self.session.backup_dump_handle))

	def restore(self, name, section="All", systemid="local"):
		"""
		Restore a backup

			>>> b.restore("mybackup.plb", "Rules & Object Configuration")

		@type name: C{str}
		@param name: Name of backup file

		@type section: C{str}
		@param section: The section in the backup file to restore,
		                such as "Rules & Object Configuration" or
		                "All" for all sections (default).

		@type systemid: C{str}
		@param systemid: Systemid to use for the items, or
		                 "local" to use system's default, or
		                 "file" to use the systemid in file.
		"""
		return plutil._do(self, self.session.backup_restore(name, section, systemid), plutil._AddCB(self))
