"""
This module contains functionality for reading "System overview" data
from the PacketLogic system.

Do NOT import this module directly. The L{PLConnection
<packetlogic2.PLConnection>} object should be used, e.g:

    >>> import packetlogic2
    >>> pl = packetlogic2.connect("192.168.1.25", "admin", "pldemo00")
    >>> so = pl.SystemOverview()

@var __revision__: Revision identifier.
"""

__revision__ = '%K%'

from resource import Resource
from thisversion import PLDB

import plutil

_Prop_SystemOverviewSystem = plutil.Prop('systemid', 'name', 'host', 'lastupdate', 'firstupdate', 'distversion', 'drdlrevision', values = plutil.RoList)

class System(object):
	""" Information about a System in the SystemOverview.
	
	Returned in a list by L{SystemOverview.system_list}.
	
	@type systemid: C{str}
	@type name: C{str}
	@type host: C{str}
	@type lastupdate: C{str}
	@type firstupdate: C{str}
	@type distversion: C{str}
	@type drdlrevision: C{str}
	@type values: read-only C{list} of C{tuple}s
	
	@cvar __slots__: For internal use only.
	"""
	__slots__ = ('_prop', '_pldb')
	
	def __init__(self, _pldb=None, **props):
		self._pldb = _pldb
		self._prop = _Prop_SystemOverviewSystem(**props)

	def __repr__(self):
		return "<%s %s @ %s '%s'>" % (self.__class__.__name__, self._prop.systemid, self._prop.host, self._prop.name)

	systemid = property(lambda self: self._prop.systemid, doc="SystemID (MachineID).")
	name = property(lambda self: self._prop.name, doc="Name given to identify this system.")
	host = property(lambda self: self._prop.host, doc="Host name or IP address.")
	lastupdate = property(lambda self: self._prop.lastupdate, doc="ISO Date/time when last updated.")
	firstupdate = property(lambda self: self._prop.firstupdate, doc="ISO Date/time when first updated.")
	distversion = property(lambda self: self._prop.distversion, doc="Version of distribution/firmware.")
	drdlrevision = property(lambda self: self._prop.drdlrevision, doc="DRDL Revision.")
	values = property(lambda self: self._prop.values, doc="Read-only C{list} of two-item C{tuple} values.")

# At the moment there is just one signel flag, but since it is a bitmap, we prepare for the future with a Flags object.
_Prop_TemplateFlags = plutil.Prop(summary=bool)

class TemplateFlags(plutil.BaseFlags):
	__slots__ = ('_prop')
	_Prop_class = _Prop_TemplateFlags
	
	def _set_flags_other_type(self, flags):
		if isinstance(flags, (int, long)):
			self._prop.update(summary=flags & PLDB.SYSTEM_OVERVIEW_FLAGS_SUMMARY)
		else:
			plutil.BaseFlags._set_flags_other_type(self, flags)
		
	#def __int__(self):
	#	return (self._prop.summary * PLDB.SYSTEM_OVERVIEW_FLAGS_SUMMARY)
	
	summary = property(lambda self: self._prop.summary)

_Prop_SystemOverviewTemplate = plutil.Prop('templateid', 'name', 'type', 'vtype', 'zone', 'id', 'ttype', flags=TemplateFlags)

class Template(object):
	""" Information about a Template in the SystemOverview.
	
	Returned in a list by L{SystemOverview.template_list}.
	
	@type templateid: C{int}
	@type name: C{str}
	@type type: C{int}
	@type flags: L{Flags <plutil.Flags>} object
	@type vtype: C{int}
	@type zone: C{int}
	@type id: C{int}
	@type ttype: C{int}
	
	@cvar __slots__: For internal use only.
	"""
	__slots__ = ('_prop', '_pldb')
	
	def __init__(self, _pldb=None, **props):
		self._pldb = _pldb
		self._prop = _Prop_SystemOverviewTemplate(**props)

	def __repr__(self):
		return "<%s #%d '%s'>" % (self.__class__.__name__, self._prop.templateid, self._prop.name)

	templateid = property(lambda self: self._prop.templateid, doc="Internal ID, used in L{SystemOverview.system_graph} call")
	name = property(lambda self: self._prop.name, doc="Name of the template.")
	type = property(lambda self: self._prop.type, doc="one of the TYPE_ constants")
	flags = property(lambda self: self._prop.flags, doc="Boolean flag(s): summary")
	vtype = property(lambda self: self._prop.vtype, doc="one of the VTYPE_ constants")
	zone = property(lambda self: self._prop.zone, doc="Zone ID, one of the STATS_ZONE_ constants")
	id = property(lambda self: self._prop.id, doc="Values internal ID within the zone.")
	ttype = property(lambda self: self._prop.ttype, doc="one of the TEMPLATE_TYPE_ constants")

class SystemOverview(Resource):
	""" This is the resource for System Overview.

	@cvar TYPE_STRING: ?
	@cvar TYPE_INT: Integer ?
	@cvar TYPE_RATE: Integer ?
	@cvar TYPE_BANDWIDTH: Integer ?
	@cvar TYPE_TRANSFER: Integer ?
	@cvar TYPE_DURATION: ?
	@cvar TYPE_TIMESTAMP: ?
	
	@cvar VTYPE_TOTAL: Value is a total since system restart.
	@cvar VTYPE_RATE: Value is a speed or rate.
	@cvar VTYPE_MAX: Value is the maximum reached since system restart.
	
	@cvar TEMPLATE_TYPE_STATS: Value is from System Statistics.
	@cvar TEMPLATE_TYPE_CHANNEL: Value is from Channel Statistics.
	
	@cvar STATS_ZONE_: All Stats Zones enumerated.
	
	"""
	
	TYPE_STRING = PLDB.SYSTEM_OVERVIEW_TYPE_STRING
	TYPE_INT = PLDB.SYSTEM_OVERVIEW_TYPE_INT
	TYPE_RATE = PLDB.SYSTEM_OVERVIEW_TYPE_RATE
	TYPE_BANDWIDTH = PLDB.SYSTEM_OVERVIEW_TYPE_BANDWIDTH
	TYPE_TRANSFER = PLDB.SYSTEM_OVERVIEW_TYPE_TRANSFER
	TYPE_DURATION = PLDB.SYSTEM_OVERVIEW_TYPE_DURATION
	TYPE_TIMESTAMP = PLDB.SYSTEM_OVERVIEW_TYPE_TIMESTAMP
	VTYPE_TOTAL = PLDB.SYSTEM_OVERVIEW_VTYPE_TOTAL
	VTYPE_RATE = PLDB.SYSTEM_OVERVIEW_VTYPE_RATE
	VTYPE_MAX = PLDB.SYSTEM_OVERVIEW_VTYPE_MAX
	TEMPLATE_TYPE_STATS = PLDB.SYSTEM_OVERVIEW_TEMPLATE_TYPE_STATS
	TEMPLATE_TYPE_CHANNEL = PLDB.SYSTEM_OVERVIEW_TEMPLATE_TYPE_CHANNEL
	STATS_ZONE_INTERFACE = PLDB.STATS_ZONE_INTERFACE
	STATS_ZONE_MEMORY = PLDB.STATS_ZONE_MEMORY
	STATS_ZONE_LINKLEVEL = PLDB.STATS_ZONE_LINKLEVEL
	STATS_ZONE_IPV4 = PLDB.STATS_ZONE_IPV4
	STATS_ZONE_HOST = PLDB.STATS_ZONE_HOST
	STATS_ZONE_TCPV4 = PLDB.STATS_ZONE_TCPV4
	STATS_ZONE_CONNECTION = PLDB.STATS_ZONE_CONNECTION
	STATS_ZONE_RULESET = PLDB.STATS_ZONE_RULESET
	STATS_ZONE_FW = PLDB.STATS_ZONE_FW
	STATS_ZONE_SHAPING = PLDB.STATS_ZONE_SHAPING
	STATS_ZONE_LIST = PLDB.STATS_ZONE_LIST
	STATS_ZONE_HASH = PLDB.STATS_ZONE_HASH
	STATS_ZONE_PPPOE = PLDB.STATS_ZONE_PPPOE
	STATS_ZONE_GENERAL = PLDB.STATS_ZONE_GENERAL
	STATS_ZONE_USER_BGP = PLDB.STATS_ZONE_USER_BGP
	STATS_ZONE_USER_MEMORY = PLDB.STATS_ZONE_USER_MEMORY
	STATS_ZONE_USER_LIST = PLDB.STATS_ZONE_USER_LIST
	STATS_ZONE_USER_HASH = PLDB.STATS_ZONE_USER_HASH
	STATS_ZONE_USER_REAPER_CMD = PLDB.STATS_ZONE_USER_REAPER_CMD
	STATS_ZONE_DRDL = PLDB.STATS_ZONE_DRDL
	STATS_ZONE_ETHERNET = PLDB.STATS_ZONE_ETHERNET
	
	def _resource(self):
		return PLDB.RESOURCE_SYSTEM_OVERVIEW

	def template_list(self):
		""" Return templates used by systems overview.
		
		@returns: C{list} of L{Template} objects.
		"""
		
		return plutil._do(self, self.session.system_overview_template_list(),
				  plutil._ListCB(self, self.session.system_overview_template_list_handle, Template))

	def system_list(self):
		""" Return systems defined in systems overview.
		
		@returns: C{list} of L{System} objects.
		"""
		
		return plutil._do(self, self.session.system_overview_list(),
				  plutil._ListCB(self, self.session.system_overview_list_handle, System))

	def system_graph(self, templateid, systemid=None):
		""" Return a C{list} of graph data for the specified system.
		Each value represent one minute, and there are 1439 values (24 hours
		except the minute in progress). Each value is an C{int} with the
		value, or C{False} if data was missing at that point.

		@type templateid: C{int}
		@param templateid: The templateid to use, see L{template_list}.
		
		@type systemid: C{str}
		@param systemid: The systemid for which to fetch data.
				 See L{system_list}. Defaults to the connected system.

		@returns: C{list} of C{int} or {False}

		@raises ValueError: if no data can be read for the specified criteria.
		"""
		if systemid is None:
			systemid = self.systemid
			
		lst = plutil._do(self, self.session.system_overview_system_graph(systemid, templateid),
				  plutil._ListCB(self, self.session.system_overview_system_graph_handle, False, False))

		# emulate old behavior without timestamps
		lst = [v[0] for v in lst]

		if not [v for v in lst if v is not False]:
			# This is (most likely) an invalid systemid or templateid.
			raise ValueError, "Can't fetch information for the specified systemid and templateid"

		return lst
